/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "maya/maya.pmh"

#define FE_JSO_DEBUG			FALSE
#define FE_JSO_AUTO_DEACTIVATE	FALSE	//* WARNING self-triggering

using namespace fe;

//* static
void JointSeparateOp::changeCB(MNodeMessage::AttributeMessage a_message,
	MPlug & a_plug,MPlug & a_otherPlug,void* a_pClientData)
{
	if(a_pClientData && a_message&MNodeMessage::kAttributeSet)
	{
		JointSeparateOp* pOperator=(JointSeparateOp*)a_pClientData;
		if(!pOperator->m_deactivate)
		{
			feLog("JointSeparateOp::changeCB deactivate %p\n",pOperator);

			pOperator->dirty();
			pOperator->m_deactivate=TRUE;
		}
	}
}

//* static
void JointSeparateOp::renameCB(MObject &a_node, const MString &a_string,
	void* a_pClientData)
{
	if(a_pClientData)
	{
		JointSeparateOp* pOperator=(JointSeparateOp*)a_pClientData;
		if(!pOperator->m_dirtied)
		{
			pOperator->dirty();
			pOperator->m_dirtied=TRUE;
		}
	}
}

JointSeparateOp::~JointSeparateOp(void)
{
	cancelCallbacks();
}

void JointSeparateOp::cancelCallbacks(void)
{
	const U32 callbackCount=m_callbackIDs.size();
	for(U32 callbackIndex=0;callbackIndex<callbackCount;callbackIndex++)
	{
		MMessage::removeCallback(m_callbackIDs[callbackIndex]);
	}
	m_callbackIDs.clear();
}

void JointSeparateOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<bool>("active")=true;
	catalog<String>("active","label")="Update Joints";

#if FE_JSO_AUTO_DEACTIVATE
	catalog<String>("active","IO")="input output";
#endif

	catalog<bool>("useRef")=false;
	catalog<String>("useRef","label")="Use Reference";

	catalog<String>("root")="joints";
	catalog<String>("root","label")="Root Node Name";
	catalog<String>("root","IO")="input output";

	catalog<String>("selectJoint")="";
	catalog<String>("selectJoint","label")="Select Node Name";

	catalog< sp<Component> >("Input Surface");

	//* hook up root joint geometry attribute
	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
//	catalog<String>("Output Surface","implementation")="curve";
	catalog<bool>("Output Surface","array")=true;
}

void JointSeparateOp::handle(Record& a_rSignal)
{
	m_dirtied=FALSE;

	MObject rootObject;

	sp<SurfaceAccessibleI> spRootAccessible;
	if(accessOutput(spRootAccessible,a_rSignal))
	{
		sp<SurfaceAccessibleMaya> spSurfaceAccessibleMaya=spRootAccessible;
		rootObject=spSurfaceAccessibleMaya->meshNode();

		MFnDagNode dagNode(rootObject);
		const String dagName=dagNode.fullPathName().asChar();

//		feLog("JointSeparateOp::handle bound %d root \"%s\"\n",
//				spRootAccessible->isBound(),dagName.c_str());

		if(!dagName.empty())
		{
			catalog<String>("root")=dagName;
		}
	}

	if(m_deactivate)
	{
		catalog<bool>("active")=false;
		m_deactivate=FALSE;
	}

	if(!catalog<bool>("active"))
	{
		return;
	}

	const String rootName=catalog<String>("root");
	const String selectJoint=catalog<String>("selectJoint");
	const BWORD useRef=catalog<bool>("useRef");

#if FE_JSO_DEBUG
	feLog("JointSeparateOp::handle\n");
#endif

	String command;

	// listRelatives -allDescendents -fullPath puppet
	command.sPrintf("fe_listNodes(\"%s\")",rootName.c_str());

//	MString response;
//	MGlobal::executeCommand(command.c_str(),response,true);
	MString response=MGlobal::executeCommandStringResult(
			command.c_str(),true);

	String tokens=response.asChar();

	String selectFull;

	String token;
	while(!(token=tokens.parse()).empty())
	{
		if(token.match(".*[|]"+selectJoint+"$"))
		{
			selectFull=token;
			break;
		}
	}

//	feLog("JointSeparateOp::handle selectJoint \"%s\"\n  \"%s\"\n  \"%s\"\n",
//			selectJoint.c_str(),m_selectedJoint.c_str(),selectFull.c_str());

	if(!selectFull.empty() && selectFull!=m_selectedJoint)
	{
		MGlobal::selectByName(selectFull.c_str(),MGlobal::kReplaceList);
	}
	m_selectedJoint=selectFull;

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessorI> spInputName;
	if(!access(spInputName,spInputAccessible,e_primitive,"name")) return;

	sp<SurfaceAccessorI> spInputParent;
	if(!access(spInputParent,spInputAccessible,e_primitive,"parent")) return;

	sp<SurfaceAccessorI> spInputAnimX;
	if(!access(spInputAnimX,spInputAccessible,e_primitive,"animX")) return;

	sp<SurfaceAccessorI> spInputAnimY;
	if(!access(spInputAnimY,spInputAccessible,e_primitive,"animY")) return;

	sp<SurfaceAccessorI> spInputAnimZ;
	if(!access(spInputAnimZ,spInputAccessible,e_primitive,"animZ")) return;

	sp<SurfaceAccessorI> spInputAnimT;
	if(!access(spInputAnimT,spInputAccessible,e_primitive,"animT")) return;

	sp<SurfaceAccessorI> spInputRefX;
	if(!access(spInputRefX,spInputAccessible,e_primitive,"refX")) return;

	sp<SurfaceAccessorI> spInputRefY;
	if(!access(spInputRefY,spInputAccessible,e_primitive,"refY")) return;

	sp<SurfaceAccessorI> spInputRefZ;
	if(!access(spInputRefZ,spInputAccessible,e_primitive,"refZ")) return;

	sp<SurfaceAccessorI> spInputRefT;
	if(!access(spInputRefT,spInputAccessible,e_primitive,"refT")) return;

	MStatus status;

	String sourceName;
	std::vector<String>* pJointNameArray=NULL;

	//* snoop on the input node
	sp<SurfaceAccessibleMaya> spSurfaceAccessibleMaya=spInputAccessible;
	if(spSurfaceAccessibleMaya.isValid())
	{
		MObject meshNode=spSurfaceAccessibleMaya->meshNode();
		MFnDependencyNode dependencyNode(meshNode,&status);
		if(status)
		{
			MPxNode *pNode=dependencyNode.userNode(&status);
			if(status)
			{
				MayaNode* pMayaNode=dynamic_cast<MayaNode*>(pNode);
				if(pMayaNode)
				{
					sp<OperatorSurfaceI> spOperator=
							pMayaNode->operatorSurface();
					sp<Catalog> spCatalog=spOperator;
					if(spCatalog.isValid() &&
							spCatalog->cataloged("jointName") &&
							spCatalog->cataloged("jointStretch") &&
							spCatalog->cataloged("jointBend") &&
							spCatalog->cataloged("jointForward") &&
							spCatalog->cataloged("jointBack"))
					{
						sourceName=dependencyNode.name().asChar();
						pJointNameArray=
								&spCatalog->catalog< std::vector<String> >(
								"jointName");
					}
				}
			}
		}
	}

	MSelectionList originalList;
	MGlobal::getActiveSelectionList(originalList);

	if(rootObject.isNull())
	{
		//* TODO is there an easier way to get a node object by name?
		MGlobal::selectByName(rootName.c_str(),MGlobal::kReplaceList);

		MSelectionList selectionList;
		MGlobal::getActiveSelectionList(selectionList);

		if(selectionList.length())
		{
			selectionList.getDependNode(0,rootObject);
		}
		if(rootObject.isNull())
		{
			MFnTransform fnTransform;
			rootObject=fnTransform.create(MObject::kNullObj,&status);
			if(!status)
			{
				feLog("JointSeparateOp::handle create root failed: %s\n",
						status.errorString().asChar());

				MGlobal::setActiveSelectionList(originalList);
				return;
			}
			fnTransform.setName(rootName.c_str());
		}
	}

	//* also assures UI refreshes
	MGlobal::setActiveSelectionList(originalList);

#if FE_JSO_DEBUG
	MFnDagNode dagRoot(rootObject);
	feLog("JointSeparateOp::handle root \"%s\" dag \"%s\" \"%s\" \"%s\"\n",
			rootName.c_str(),
			dagRoot.name().asChar(),
			dagRoot.partialPathName().asChar(),
			dagRoot.fullPathName().asChar());
#endif

	cancelCallbacks();

	const U32 primitiveCount=spInputName->count();

#if FE_JSO_DEBUG
	feLog("JointSeparateOp::handle primitiveCount %d\n",primitiveCount);
#endif

	std::map<String,MObject> objectMap;

	m_keepList.clear();

	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
#if FE_JSO_DEBUG
		feLog("JointSeparateOp::handle %d/%d\n",primitiveIndex,primitiveCount);
#endif

		const String name=spInputName->string(primitiveIndex);

		const String parentName=spInputParent->string(primitiveIndex);
		MObject parentObject=objectMap[parentName];
		if(parentObject==MObject::kNullObj)
		{
			parentObject=rootObject;
		}

		//* watch for any manual change on each output joint
		void* pClientData=this;
		MCallbackId callbackID=MNodeMessage::addNameChangedCallback(
				parentObject,renameCB,pClientData,&status);
		m_callbackIDs.push_back(callbackID);
#if FE_JSO_AUTO_DEACTIVATE
		callbackID=MNodeMessage::addAttributeChangedCallback(
				parentObject,changeCB,pClientData,&status);
		m_callbackIDs.push_back(callbackID);
#endif

		SpatialTransform xformLocalRef;
		xformLocalRef.column(0)=spInputRefX->spatialVector(primitiveIndex);
		xformLocalRef.column(1)=spInputRefY->spatialVector(primitiveIndex);
		xformLocalRef.column(2)=spInputRefZ->spatialVector(primitiveIndex);
		xformLocalRef.column(3)=spInputRefT->spatialVector(primitiveIndex);

		SpatialTransform xformLocal;
		if(useRef)
		{
			xformLocal=xformLocalRef;
		}
		else
		{
			xformLocal.column(0)=spInputAnimX->spatialVector(primitiveIndex);
			xformLocal.column(1)=spInputAnimY->spatialVector(primitiveIndex);
			xformLocal.column(2)=spInputAnimZ->spatialVector(primitiveIndex);
			xformLocal.column(3)=spInputAnimT->spatialVector(primitiveIndex);
		}

		SpatialTransform invLocalRef;
		invert(invLocalRef,xformLocalRef);

		const SpatialVector localDisplace=xformLocal.translation();

		const SpatialQuaternion quatRotation=xformLocal*invLocalRef;
		const SpatialQuaternion quatOrientation=xformLocalRef;

		const MVector mDisplace(
				localDisplace[0],localDisplace[1],localDisplace[2]);
		const MQuaternion mRotation(
				quatRotation[0],quatRotation[1],
				quatRotation[2],quatRotation[3]);
		const MQuaternion mOrientation(
				quatOrientation[0],quatOrientation[1],
				quatOrientation[2],quatOrientation[3]);

#if FE_JSO_DEBUG
		feLog("  \"%s\" parent \"%s\"\n",name.c_str(),parentName.c_str());
		feLog("  point %s localDisplace %s\n",
				c_print(xformLocal.translation()),c_print(localDisplace));
		feLog("  quatRotation %s\n",c_print(quatRotation));
		feLog("  quatOrientation %s\n",c_print(quatOrientation));
#endif

		MObject jointObject;

		MFnDagNode fnDagNode(parentObject);
		const U32 childCount=fnDagNode.childCount(&status);
		for(U32 childIndex=0;childIndex<childCount;childIndex++)
		{
			MObject child=fnDagNode.child(childIndex,&status);
			MFnDagNode dagChild(child);
			if(dagChild.name()==name.c_str())
			{
				jointObject=child;
				break;
			}
		}

		MFnIkJoint fnJoint(jointObject);

		if(jointObject.isNull())
		{
			jointObject=fnJoint.create(parentObject,&status);
			if(!status)
			{
				feLog("JointSeparateOp::handle create failed: %s\n",
						status.errorString().asChar());
				continue;
			}

			fnJoint.setName(name.c_str());

			MFnNumericAttribute numAttr;

			const char* stretchName="stretch";
			MObject stretch=numAttr.create(stretchName,stretchName,
					MFnNumericData::k3Double);
			numAttr.setKeyable(true);
			numAttr.setDefault(1.0,1.0,1.0);
			fnJoint.addAttribute(stretch);

			const char* bendName="bend";
			MObject bend=numAttr.create(bendName,bendName,
					MFnNumericData::k3Double);
			numAttr.setKeyable(true);
			numAttr.setDefault(0.0,0.0,0.0);
			fnJoint.addAttribute(bend);

			const char* forwardName="forward";
			MObject forward=numAttr.create(forwardName,forwardName,
					MFnNumericData::k3Double);
			numAttr.setKeyable(true);
			numAttr.setDefault(180.0,180.0,180.0);
			fnJoint.addAttribute(forward);

			const char* backName="back";
			MObject back=numAttr.create(backName,backName,
					MFnNumericData::k3Double);
			numAttr.setKeyable(true);
			numAttr.setDefault(180.0,180.0,180.0);
			fnJoint.addAttribute(back);
		}

		if(pJointNameArray)
		{
			std::vector<String>& rJointNameArray=*pJointNameArray;

			const U32 nameCount=rJointNameArray.size();
			for(U32 nameIndex=0;nameIndex<nameCount;nameIndex++)
			{
				if(rJointNameArray[nameIndex]==name)
				{
					//* TODO full joint path

					MString response;
					String command;

					// connectAttr -f Joint.stretch PuppetOp.jointStretch[1]
					command.sPrintf("connectAttr -f %s.stretch"
							" %s.jointStretch[%d]",
							name.c_str(),sourceName.c_str(),nameIndex+1);

					MGlobal::executeCommand(command.c_str(),response);

					// connectAttr -f Joint.bend PuppetOp.jointBend[1]
					command.sPrintf("connectAttr -f %s.bend"
							" %s.jointBend[%d]",
							name.c_str(),sourceName.c_str(),nameIndex+1);

					MGlobal::executeCommand(command.c_str(),response);

					// connectAttr -f Joint.forward PuppetOp.jointForward[1]
					command.sPrintf("connectAttr -f %s.forward"
							" %s.jointForward[%d]",
							name.c_str(),sourceName.c_str(),nameIndex+1);

					MGlobal::executeCommand(command.c_str(),response);

					// connectAttr -f Joint.back PuppetOp.jointBack[1]
					command.sPrintf("connectAttr -f %s.back"
							" %s.jointBack[%d]",
							name.c_str(),sourceName.c_str(),nameIndex+1);

					MGlobal::executeCommand(command.c_str(),response);
				}
			}
		}

		objectMap[name]=jointObject;
		m_keepList.push_back(jointObject);

		fnJoint.setTranslation(mDisplace,MSpace::kTransform);
		fnJoint.setRotation(mRotation,MSpace::kTransform);
		fnJoint.setOrientation(mOrientation);
	}

	//* delete unused pre-existing joints
	scanDag(rootObject);

#if FE_JSO_DEBUG
	feLog("JointSeparateOp::handle done\n");
#endif
}

void JointSeparateOp::scanDag(MObject a_parent)
{
	if(a_parent.isNull())
	{
		return;
	}

	MFnDagNode dagParent(a_parent);

	std::vector<MObject> deleteList;

	MStatus status;
	const U32 childCount=dagParent.childCount(&status);
	for(U32 childIndex=0;childIndex<childCount;childIndex++)
	{
		MObject child=dagParent.child(childIndex,&status);

		const U32 keepCount=m_keepList.size();
		U32 keepIndex=0;
		for(;keepIndex<keepCount;keepIndex++)
		{
			if(m_keepList[keepIndex]==child)
			{
				scanDag(child);
				break;
			}
		}
		if(keepIndex==keepCount)
		{
			deleteList.push_back(child);
		}
	}

	const U32 deleteCount=deleteList.size();
	for(U32 deleteIndex=0;deleteIndex<deleteCount;deleteIndex++)
	{
		MObject child=deleteList[deleteIndex];

		MGlobal::deleteNode(child);
	}
}
